@@ -46,10 +46,10 @@ def merge_ranges(ranges):
             merged_ranges.append([start, end])
     return merged_ranges
 
-def init_regions_for(args):
+def init_regions_for(args_regions, args_ops, args_numa_node):
     init_regions = []
-    if args.regions:
-        for region in args.regions.split():
+    if args_regions:
+        for region in args_regions.split():
             addrs = region.split('-')
             try:
                 if len(addrs) != 2:
@@ -63,10 +63,10 @@ def init_regions_for(args):
                 return None, 'Wrong \'--regions\' argument (%s)' % e
             init_regions.append(region)
 
-    if args.ops == 'paddr' and not init_regions:
-        if args.numa_node != None:
+    if args_ops == 'paddr' and not init_regions:
+        if args_numa_node != None:
             init_regions, err = damo_pa_layout.numa_addr_ranges(
-                    args.numa_node)
+                    args_numa_node)
             if err != None:
                 return None, err
             init_regions = merge_ranges(init_regions)
@@ -87,19 +87,23 @@ def override_vals(to_override, new_vals):
         if new_val is not None:
             to_override[idx] = new_val
 
-def damon_intervals_for(args):
+def damon_intervals_for(args_intervals, args_sample, args_aggr, args_updr,
+                        args_intervals_goal):
     intervals = ['5ms', '100ms', '1s']
-    override_vals(intervals, args.monitoring_intervals)
-    override_vals(intervals, [args.sample, args.aggr, args.updr])
+    override_vals(intervals, args_intervals)
+    override_vals(intervals, [args_sample, args_aggr, args_updr])
 
-    intervals_goal = _damon.DamonIntervalsGoal(*args.monitoring_intervals_goal)
+    if args_intervals_goal is None:
+        args_intervals_goal = ['0%', '0', '0us', '0us']
+
+    intervals_goal = _damon.DamonIntervalsGoal(*args_intervals_goal)
 
     return _damon.DamonIntervals(*intervals, intervals_goal)
 
-def damon_nr_regions_range_for(args):
+def damon_nr_regions_range_for(args_range, args_minr, args_maxr):
     nr_range = ['10', '1000']
-    override_vals(nr_range, args.monitoring_nr_regions_range)
-    override_vals(nr_range, [args.minr, args.maxr])
+    override_vals(nr_range, args_range)
+    override_vals(nr_range, [args_minr, args_maxr])
     return _damon.DamonNrRegionsRange(*nr_range)
 
 def schemes_option_to_damos(schemes):
@@ -431,37 +435,169 @@ def damos_for(args):
         return None, 'failed damo schemes arguments parsing (%s)' % err
     return schemes, None
 
-def damon_ctx_for(args):
-    try:
-        intervals = damon_intervals_for(args)
-    except Exception as e:
-        return None, 'invalid intervals arguments (%s)' % e
-    try:
-        nr_regions = damon_nr_regions_range_for(args)
-    except Exception as e:
-        return None, 'invalid nr_regions arguments (%s)' % e
-    ops = args.ops
-
-    init_regions, err = init_regions_for(args)
+def damon_target_for(args, idx, ops):
+    init_regions, err = init_regions_for(
+            args.regions[idx], ops, args.numa_node[idx])
     if err:
         return None, err
 
     try:
-        target = _damon.DamonTarget(args.target_pid
+        target = _damon.DamonTarget(args.target_pid[idx]
                 if _damon.target_has_pid(ops) else None, init_regions)
     except Exception as e:
         return 'Wrong \'--target_pid\' argument (%s)' % e
+    return target, None
 
-    schemes, err = damos_for(args)
-    if err:
-        return None, err
+def damon_ctx_for(args, idx):
+    if args.ops[idx] is None:
+        if args.target_pid[idx] is None:
+            args.ops[idx] = 'paddr'
+        else:
+            args.ops[idx] = 'vaddr'
 
     try:
-        ctx = _damon.DamonCtx(ops, [target], intervals, nr_regions, schemes)
+        intervals = damon_intervals_for(
+                args.monitoring_intervals[idx], args.sample[idx],
+                args.aggr[idx], args.updr[idx],
+                args.monitoring_intervals_goal[idx])
+    except Exception as e:
+        return None, 'invalid intervals arguments (%s)' % e
+    try:
+        nr_regions = damon_nr_regions_range_for(
+                args.monitoring_nr_regions_range[idx],
+                args.minr[idx], args.maxr[idx])
+    except Exception as e:
+        return None, 'invalid nr_regions arguments (%s)' % e
+    ops = args.ops[idx]
+
+    try:
+        ctx = _damon.DamonCtx(ops, None, intervals, nr_regions, schemes=[])
         return ctx, None
     except Exception as e:
         return None, 'Creating context from arguments failed (%s)' % e
 
+def get_nr_ctxs(args):
+    candidates = []
+    for v in [args.ops, args.sample, args.aggr, args.updr, args.minr,
+              args.maxr, args.monitoring_intervals,
+              args.monitoring_intervals_goal,
+              args.monitoring_nr_regions_range]:
+        if v is not None:
+            candidates.append(len(v))
+    if args.nr_targets is not None:
+        candidates.append(len(args.nr_targets))
+    if args.nr_schemes is not None:
+        candidates.append(len(args.nr_schemes))
+    if len(candidates) == 0:
+        return 1
+    return max(candidates)
+
+def get_nr_targets(args):
+    candidates = []
+    for v in [args.target_pid, args.regions, args.numa_node]:
+        if v is not None:
+            candidates.append(len(v))
+    candidates.append(get_nr_ctxs(args))
+    return max(candidates)
+
+def fillup_none_ctx_args(args):
+    nr_ctxs = get_nr_ctxs(args)
+    for attr_name in [
+            'ops', 'sample', 'aggr', 'updr', 'minr', 'maxr',
+            'monitoring_intervals', 'monitoring_intervals_goal',
+            'monitoring_nr_regions_range']:
+        attr_val = getattr(args, attr_name)
+        if attr_val is None:
+            setattr(args, attr_name, [None] * nr_ctxs)
+        elif len(attr_val) < nr_ctxs:
+            print(attr_name, attr_val)
+            setattr(args, attr_name,
+                    attr_val + [None] * (nr_ctxs - len(attr_val)))
+
+def fillup_none_target_args(args):
+    nr_targets = get_nr_targets(args)
+    for attr_name in ['target_pid', 'regions', 'numa_node']:
+        attr_val = getattr(args, attr_name)
+        if attr_val is None:
+            setattr(args, attr_name, [None] * nr_targets)
+        elif len(attr_val) < nr_targets:
+            print(attr_name, attr_val)
+            setattr(args, attr_name,
+                    attr_val + [None] * (nr_targets - len(attr_val)))
+
+def gen_assign_targets(ctxs, args):
+    nr_targets = get_nr_targets(args)
+    targets = []
+    if args.nr_targets is None:
+        if len(ctxs) != 1:
+            return '--nr_targets is required for multiple contexts'
+        args.nr_targets = [nr_targets]
+    for idx in range(nr_targets):
+        for i in range(len(args.nr_targets)):
+            if idx < sum(args.nr_targets[:i + 1]):
+                ops = args.ops[i]
+                break
+        target, err = damon_target_for(args, idx, ops)
+        if err is not None:
+            return None, err
+        targets.append(target)
+    if sum(args.nr_targets) != len(targets):
+        return '--nr_targets and number of targets mismatch (%d != %d)' % (
+                sum(args.nr_targets), len(targets))
+    if len(args.nr_targets) != len(ctxs):
+        return '--nr_targets and number of ctxs mismatch (%d != %d)' % (
+                len(args.nr_targets), len(ctxs))
+    ctx_idx = 0
+    target_idx = 0
+    for nr in args.nr_targets:
+        ctxs[ctx_idx].targets = targets[target_idx:target_idx + nr]
+        ctx_idx += 1
+        target_idx += nr
+    return None
+
+def gen_assign_schemes(ctxs, args):
+    schemes, err = damos_for(args)
+    if err is not None:
+        return err
+    if args.nr_schemes is None:
+        if len(ctxs) != 1 and len(schemes) > 0:
+            return '--nr_schemes is required for multiple contexts'
+        args.nr_schemes = [len(schemes)]
+        args.nr_schemes += [0] * (len(ctxs) - 1)
+    if sum(args.nr_schemes) != len(schemes):
+        return '--nr_schemes and number of schemes mismatch (%d != %d)' % (
+                sum(args.nr_schemes), len(schemes))
+    if len(args.nr_schemes) != len(ctxs):
+        return '--nr_schemes and number of ctxs mismatch (%d != %d)' % (
+                len(args.nr_schemes), len(ctxs))
+    ctx_idx = 0
+    scheme_idx = 0
+    for nr in args.nr_schemes:
+        ctxs[ctx_idx].schemes = schemes[scheme_idx:scheme_idx + nr]
+        ctx_idx += 1
+        scheme_idx += nr
+    return None
+
+def damon_ctxs_for(args):
+    fillup_none_ctx_args(args)
+    fillup_none_target_args(args)
+    ctxs = []
+    for idx in range(get_nr_ctxs(args)):
+        ctx, err = damon_ctx_for(args, idx)
+        if err is not None:
+            return None, err
+        ctxs.append(ctx)
+
+    err = gen_assign_targets(ctxs, args)
+    if err is not None:
+        return None, err
+
+    err = gen_assign_schemes(ctxs, args)
+    if err is not None:
+        return None, err
+
+    return ctxs, err
+
 def kdamonds_from_json_arg(arg):
     try:
         if os.path.isfile(arg):
@@ -509,14 +645,15 @@ def warn_option_override(option_name):
     print('warning: %s is overridden by <deducible target>' % option_name)
 
 def deduce_target_update_args(args):
+    'deducible target supports only single context/single kdamond'
     args.self_started_target = False
     target_type = deduced_target_type(args.deducible_target)
     if target_type == target_type_unknown:
         return 'target \'%s\' is not supported' % args.deducible_target
     if target_type == target_type_explicit and args.deducible_target == 'paddr':
-        if not args.ops in ['paddr', None]:
+        if not args.ops in [['paddr'], None]:
             warn_option_override('--ops')
-        args.ops = 'paddr'
+        args.ops = ['paddr']
         if args.target_pid != None:
             warn_option_override('--target_pid')
         args.target_pid = None
@@ -530,15 +667,15 @@ def deduce_target_update_args(args):
         pid = int(args.deducible_target)
     if args.target_pid != None:
         print('warning: --target_pid will be ignored')
-    args.target_pid = pid
+    args.target_pid = [pid]
     if not args.regions:
-        if not args.ops in ['vaddr', None]:
+        if not args.ops in [['vaddr'], None]:
             warn_option_override('--ops')
-        args.ops = 'vaddr'
+        args.ops = ['vaddr']
     if args.regions:
-        if not args.ops in ['fvaddr', None]:
+        if not args.ops in [['fvaddr'], None]:
             print('warning: override --ops by <deducible target> and --regions')
-        args.ops = 'fvaddr'
+        args.ops = ['fvaddr']
 
 def evaluate_args(args):
     '''
@@ -587,16 +724,27 @@ def kdamonds_for(args):
         err = deduce_target_update_args(args)
         if err:
             return None, err
-    if args.ops == None:
-        if args.target_pid == None:
-            args.ops = 'paddr'
-        else:
-            args.ops = 'vaddr'
 
-    ctx, err = damon_ctx_for(args)
+    ctxs, err = damon_ctxs_for(args)
     if err:
         return None, err
-    return [_damon.Kdamond(state=None, pid=None, contexts=[ctx])], None
+
+    if args.nr_ctxs is None:
+        args.nr_ctxs = [len(ctxs)]
+    if sum(args.nr_ctxs) != len(ctxs):
+        return None, '--nr_ctxs and number of ctxs mismatch (%d != %d)' % (
+                sum(args.nr_ctxs), len(ctxs))
+
+    kdamonds = []
+    ctx_idx = 0
+    for nr in args.nr_ctxs:
+        try:
+            kdamonds.append(_damon.Kdamond(
+                state=None, pid=None, contexts=ctxs[ctx_idx:ctx_idx + nr]))
+        except Exception as e:
+            return None, 'kdamond creation fail (%s)' % e
+        ctx_idx += nr
+    return kdamonds, None
 
 def self_started_target(args):
     return 'self_started_target' in args and args.self_started_target
@@ -644,59 +792,69 @@ def set_common_argparser(parser):
 def set_monitoring_attrs_pinpoint_argparser(parser, hide_help=False):
     # for easier pinpoint setup
     parser.add_argument(
-            '-s', '--sample', metavar='<microseconds>',
+            '-s', '--sample', metavar='<microseconds>', action='append',
             help='sampling interval (us)'
             if not hide_help else argparse.SUPPRESS)
     parser.add_argument(
-            '-a', '--aggr', metavar='<microseconds>',
+            '-a', '--aggr', metavar='<microseconds>', action='append',
             help='aggregate interval (us)'
             if not hide_help else argparse.SUPPRESS)
     parser.add_argument(
-            '-u', '--updr', metavar='<microseconds>',
+            '-u', '--updr', metavar='<microseconds>', action='append',
             help='regions update interval (us)'
             if not hide_help else argparse.SUPPRESS)
     parser.add_argument(
-            '-n', '--minr', metavar='<# regions>',
+            '-n', '--minr', metavar='<# regions>', action='append',
             help='minimal number of regions'
             if not hide_help else argparse.SUPPRESS)
     parser.add_argument(
-            '-m', '--maxr', metavar='<# regions>',
+            '-m', '--maxr', metavar='<# regions>', action='append',
             help='maximum number of regions'
             if not hide_help else argparse.SUPPRESS)
 
 def set_monitoring_attrs_argparser(parser, hide_help=False):
     # for easier total setup
-    parser.add_argument('--monitoring_intervals', nargs=3,
+    parser.add_argument('--monitoring_intervals', nargs=3, action='append',
                         metavar=('<sample>', '<aggr>', '<update>'),
                         help='monitoring intervals (us)'
                         if not hide_help else argparse.SUPPRESS)
     parser.add_argument(
-            '--monitoring_intervals_goal', nargs=4,
+            '--monitoring_intervals_goal', nargs=4, action='append',
             metavar=('<access_bp>', '<aggrs>', '<min_sample_us>',
-                     '<max_sample_us>'), default=['0%', '0', '0us', '0us'],
+                     '<max_sample_us>'),
             help='monitoring intervals auto-tuning goal'
             if not hide_help else argparse.SUPPRESS)
     parser.add_argument('--monitoring_nr_regions_range', nargs=2,
-                        metavar=('<min>', '<max>'),
+                        metavar=('<min>', '<max>'), action='append',
                         help='min/max number of monitoring regions'
                         if not hide_help else argparse.SUPPRESS)
 
 def set_monitoring_argparser(parser, hide_help=False):
     parser.add_argument('--ops', choices=['vaddr', 'paddr', 'fvaddr'],
+                        action='append',
                         help='monitoring operations set'
                         if not hide_help else argparse.SUPPRESS)
     parser.add_argument('--target_pid', type=int, metavar='<pid>',
+                        action='append',
                         help='pid of monitoring target process'
                         if not hide_help else argparse.SUPPRESS)
     parser.add_argument('-r', '--regions', metavar='"<start>-<end> ..."',
-                        type=str, default='',
+                        type=str, action='append',
                         help='monitoring target address regions'
                         if not hide_help else argparse.SUPPRESS)
     parser.add_argument(
             '--numa_node', metavar='<node id>', type=int, nargs='+',
+            action='append',
             help='if target is \'paddr\', limit it to the numa node'
             if not hide_help else argparse.SUPPRESS)
     set_monitoring_attrs_argparser(parser, hide_help)
+    parser.add_argument(
+            '--nr_targets', nargs='+', type=int,
+            help='number of monitoring targets for each context (in order)'
+            if not hide_help else argparse.SUPPRESS)
+    parser.add_argument('--nr_ctxs', nargs='+', type=int,
+                        help='number of contexts for each kdamond (in order)'
+                        if not hide_help else argparse.SUPPRESS)
 
 def set_damos_argparser(parser, hide_help):
     parser.add_argument('--damos_action', metavar='<action>', nargs='+',
@@ -780,6 +938,9 @@ def set_damos_argparser(parser, hide_help):
                 '<low mark (permil)>'),
             help='damos watermarks'
             if not hide_help else argparse.SUPPRESS)
+    parser.add_argument('--nr_schemes', nargs='+', type=int,
+                        help='number of schemes for each context (in order)'
+                        if not hide_help else argparse.SUPPRESS)
 
 def set_misc_damon_params_argparser(parser):
     parser.add_argument('-c', '--schemes', metavar='<json string or file>',
