@@ -52,6 +52,13 @@ record_formatters = [
                 record.snapshots[-1].end_time - record.snapshots[0].start_time,
                 fmt.raw_number),
             'duration of the record'),
+        Formatter('<intervals>',
+                  lambda record, fmt: record_intervals(record, fmt.raw_number),
+                  'monitoring intervals'),
+        Formatter('<intervals goal>',
+                  lambda record, fmt:
+                  record.intervals.intervals_goal.to_str(fmt.raw_number),
+                  'monitoring intervals'),
         Formatter('<format strings>',
                   lambda record, fmt: format_strings(fmt),
                   'current format strings')
@@ -156,6 +163,11 @@ snapshot_formatters = [
                 lambda snapshot, record, fmt:
                 intervals_tuning_status(snapshot, record, fmt),
                 'intervals tuning status'),
+        Formatter(
+                '<damos stats>',
+                lambda snapshot, record, fmt:
+                damos_stats_str(snapshot, record, fmt),
+                'DAMOS stats for the snapshot-retrieval scheme'),
         ]
 
 region_formatters = [
@@ -215,6 +227,12 @@ default_region_format = '<index> addr <start address> size <size> access <access
 default_snapshot_tail_format = 'memory bw estimate: <estimated memory bandwidth>\ntotal size: <total bytes>'
 default_snapshot_tail_format_filter_installed = 'memory bw estimate: <estimated memory bandwidth>  df-passed: <filters passed estimated memory bandwidth>\ntotal size: <total bytes>  df-passed <filters passed bytes>'
 
+def record_intervals(record, raw_number):
+    intervals = record.intervals
+    return 'sample %s, aggr %s' % (
+            _damo_fmt_str.format_time_us(intervals.sample, raw_number),
+            _damo_fmt_str.format_time_us(intervals.aggr, raw_number))
+
 def format_strings(fmt):
     return '\n'.join([
         'record head: "%s"' % fmt.format_record_head,
@@ -251,8 +269,8 @@ def intervals_tuning_status(snapshot, record, fmt):
         max_observation += max_samples_per_region * region.size()
         observation += region.size() * region.nr_accesses.samples
     return '%s (%s / %s) %s %s' % (
-            _damo_fmt_str.format_percent(observation * 100 / max_observation,
-                                         fmt.raw_number),
+            _damo_fmt_str.format_percent(
+                observation * 100 / max(max_observation, 1), fmt.raw_number),
             _damo_fmt_str.format_sz(observation, fmt.raw_number),
             _damo_fmt_str.format_sz(max_observation, fmt.raw_number),
             _damo_fmt_str.format_time_us(record.intervals.sample,
@@ -261,6 +279,11 @@ def intervals_tuning_status(snapshot, record, fmt):
                                          fmt.raw_number),
             )
 
+def damos_stats_str(snapshot, record, fmt):
+    if snapshot.damos_stats is None:
+        return 'none'
+    return snapshot.damos_stats.to_str(fmt.raw_number)
+
 def estimated_mem_bw(snapshot, record, fmt, filter_passed_only=False):
     access_bytes = 0
     for region in snapshot.regions:
@@ -282,52 +305,45 @@ def filters_pass_type_of(record):
         text += '\n# WARN: Allow filters at the end of the list means nothing'
     return text
 
-def temperature_sz_hist_str(snapshot, record, fmt, df_passed_sz):
-    raw = fmt.raw_number
-    if len(snapshot.regions) == 0:
-        return 'no region in snapshot'
-    hist = {}
-    # set size weight zero
-    weights = [0, fmt.temperature_weights[1], fmt.temperature_weights[2]]
-    for region in snapshot.regions:
-        temperature = temperature_of(region, fmt.temperature_weights)
-        if not temperature in hist:
-            hist[temperature] = 0
-        if df_passed_sz:
-            hist[temperature] += region.sz_filter_passed
-        else:
-            hist[temperature] += region.size()
+def get_linearscale_hist_ranges(minv, maxv, nr_ranges):
+    hist_ranges = []
+    total_interval = maxv + 1 - minv
+    interval = max(math.ceil(total_interval / nr_ranges), 1)
+    for i in range(nr_ranges):
+        hist_ranges.append([minv + interval * i, minv + interval * (i + 1)])
+    return hist_ranges
 
-    temperatures = sorted(hist.keys())
-    min_temp, max_temp = temperatures[0], temperatures[-1]
-    interval = max(int((max_temp - min_temp) / 10), 1)
-    max_temp = interval * math.ceil(max_temp / interval) + 1
+def get_logscale_hist_ranges(minv, maxv, nr_ranges):
+    initial_interval = max(int((maxv - minv) / 2**nr_ranges), 1)
+    # start from 1 second interval
+    ranges = [[minv, minv + initial_interval]]
+    while ranges[-1][-1] < maxv:
+        last_interval = ranges[-1][1] - ranges[-1][0]
+        ranges.append([ranges[-1][-1], ranges[-1][-1] + last_interval * 2])
+    return ranges
 
-    hist2 = []
-    min_sz = None
-    max_sz = None
-    max_trange_str = None
-    max_sz_str = None
-    for t in range(min_temp, max_temp, interval):
-        min_t = t
-        max_t = t + interval
-        sz = sum([hist[x] for x in temperatures if x >= min_t and x < max_t])
-        if min_sz is None or sz < min_sz:
-            min_sz = sz
-        if max_sz is None or max_sz < sz:
-            max_sz = sz
-        trange_str = '[%s, %s)' % (_damo_fmt_str.format_nr(min_t, raw),
-                                   _damo_fmt_str.format_nr(max_t, raw))
-        if max_trange_str is None or max_trange_str < len(trange_str):
-            max_trange_str = len(trange_str)
-        sz_str = _damo_fmt_str.format_sz(sz, raw)
-        if max_sz_str is None or max_sz_str < len(sz_str):
-            max_sz_str = len(sz_str)
-        hist2.append([trange_str, sz_str, sz])
+def get_hist_ranges(minv, maxv, nr_ranges, logscale):
+    if logscale:
+        return get_logscale_hist_ranges(minv, maxv, nr_ranges)
+    return get_linearscale_hist_ranges(minv, maxv, nr_ranges)
+
+def histogram_str(hist):
+    """
+    Format histogram string to show.
+    Args:
+        hist (list): List of lines of histogram to show.
+    Returns:
+        str: A string of the histogram that formatted in a user-friendly way.
+    """
+    sizes = [entry[2] for entry in hist]
+    min_sz = min(sizes)
+    max_sz = max(sizes)
+    max_trange_str = max([len(entry[0]) for entry in hist])
+    max_sz_str = max([len(entry[1]) for entry in hist])
     max_dots = 20
     sz_interval = max(int((max_sz - min_sz) / max_dots), 1)
     lines = []
-    for trange_str, sz_str, sz in hist2:
+    for trange_str, sz_str, sz in hist:
         trange_str = '%s%s' % (trange_str,
                                ' ' * (max_trange_str - len(trange_str)))
         sz_str = '%s%s' % (sz_str,
@@ -338,62 +354,50 @@ def temperature_sz_hist_str(snapshot, record, fmt, df_passed_sz):
         lines.append('%s %s %s' % (trange_str, sz_str, bar))
     return '\n'.join(lines)
 
+def sz_hist_str(snapshot, fmt, df_passed_sz, get_metric_fn, fmt_metric_fn):
+    raw = fmt.raw_number
+    if len(snapshot.regions) == 0:
+        return 'no region in snapshot'
+    hist = {}
+    for region in snapshot.regions:
+        metric = get_metric_fn(region, fmt)
+        if not metric in hist:
+            hist[metric] = 0
+        if df_passed_sz:
+            hist[metric] += region.sz_filter_passed
+        else:
+            hist[metric] += region.size()
+
+    metrics = sorted(hist.keys())
+    min_metric, max_metric = metrics[0], metrics[-1]
+
+    hist2 = []
+    for min_m, max_m in get_hist_ranges(
+            min_metric, max_metric, 10, fmt.hist_logscale):
+        sz = sum([hist[x] for x in metrics if x >= min_m and x < max_m])
+        metric_range_str = '[%s, %s)' % (
+                fmt_metric_fn(min_m, raw), fmt_metric_fn(max_m, raw))
+        sz_str = _damo_fmt_str.format_sz(sz, raw)
+        hist2.append([metric_range_str, sz_str, sz])
+    return histogram_str(hist2)
+
+def temperature_sz_hist_str(snapshot, record, fmt, df_passed_sz):
+    def get_temperature(region, fmt):
+        # set size weight zero
+        weights = [0, fmt.temperature_weights[1], fmt.temperature_weights[2]]
+        return temperature_of(region, weights)
+
+    return sz_hist_str(snapshot, fmt, df_passed_sz, get_temperature,
+                       _damo_fmt_str.format_nr)
+
 def recency_hist_str(snapshot, record, fmt, df_passed_sz):
-    raw = fmt.raw_number
-    if len(snapshot.regions) == 0:
-        return 'no region in snapshot'
-    hist = {}
-    for region in snapshot.regions:
-        last_used = -1 * region.age.usec
+    def get_last_used_time(region, fmt):
         if region.nr_accesses.percent > 0:
-            last_used = 0
-        if not last_used in hist:
-            hist[last_used] = 0
-        if df_passed_sz is True:
-            hist[last_used] += region.sz_filter_passed
-        else:
-            hist[last_used] += region.size()
+            return 0
+        return region.age.usec
 
-    last_used_times = sorted(hist.keys())
-    min_lut, max_lut = last_used_times[0], last_used_times[-1]
-    interval = max(int((max_lut - min_lut) / 10), 1)
-    max_lut = interval * math.ceil(max_lut / interval) + 1
-
-    hist2 = []
-    min_sz = None
-    max_sz = None
-    max_trange_str = None
-    max_sz_str = None
-    for t in range(min_lut, max_lut, interval):
-        min_t = t
-        max_t = t + interval
-        sz = sum([hist[x] for x in last_used_times if x >= min_t and x < max_t])
-        if min_sz is None or sz < min_sz:
-            min_sz = sz
-        if max_sz is None or max_sz < sz:
-            max_sz = sz
-        trange_str = '[-%s, -%s)' % (
-                _damo_fmt_str.format_time_us(-1 * min_t, raw),
-                _damo_fmt_str.format_time_us(-1 * max_t, raw))
-        if max_trange_str is None or max_trange_str < len(trange_str):
-            max_trange_str = len(trange_str)
-        sz_str = _damo_fmt_str.format_sz(sz, raw)
-        if max_sz_str is None or max_sz_str < len(sz_str):
-            max_sz_str = len(sz_str)
-        hist2.append([trange_str, sz_str, sz])
-    max_dots = 20
-    sz_interval = max(int((max_sz - min_sz) / max_dots), 1)
-    lines = []
-    for trange_str, sz_str, sz in hist2:
-        trange_str = '%s%s' % (trange_str,
-                               ' ' * (max_trange_str - len(trange_str)))
-        sz_str = '%s%s' % (sz_str,
-                           ' ' * (max_sz_str - len(sz_str)))
-
-        nr_dots = min(math.ceil((sz - min_sz) / sz_interval), max_dots)
-        bar = '|%s%s|' % ('*' * nr_dots, ' ' * (max_dots - nr_dots))
-        lines.append('%s %s %s' % (trange_str, sz_str, bar))
-    return '\n'.join(lines)
+    return sz_hist_str(snapshot, fmt, df_passed_sz, get_last_used_time,
+                       _damo_fmt_str.format_time_us)
 
 def temperature_str(region, raw, fmt):
     temperature = temperature_of(region, fmt.temperature_weights)
@@ -932,6 +936,7 @@ class ReportFormat:
     sort_regions_dsc = None
     temperature_weights = None
     dont_merge_regions = None
+    hist_logscale = None
 
     format_record_head = None
     format_record_tail = None
@@ -966,6 +971,7 @@ class ReportFormat:
         self.sort_regions_dsc = args.sort_regions_dsc
         self.temperature_weights = args.temperature_weights
         self.dont_merge_regions = args.dont_merge_regions
+        self.hist_logscale = args.hist_logscale
         self.format_record_head = args.format_record_head
         self.format_record_tail = args.format_record_tail
         self.format_snapshot_head = args.format_snapshot_head
@@ -998,6 +1004,7 @@ class ReportFormat:
                 'sort_regions_dsc': self.sort_regions_dsc,
                 'temperature_weights': self.temperature_weights,
                 'dont_merge_regions': self.dont_merge_regions,
+                'hist_logscale': self.hist_logscale,
                 'format_record_head': self.format_record_head,
                 'format_record_tail': self.format_record_tail,
                 'format_snapshot_head': self.format_snapshot_head,
@@ -1024,6 +1031,11 @@ class ReportFormat:
         self.sort_regions_dsc = kvpairs['sort_regions_dsc']
         self.temperature_weights = kvpairs['temperature_weights']
         self.dont_merge_regions = kvpairs['dont_merge_regions']
+        # hist_logscale introduced after v2.7.0
+        if 'hist_logscale' in kvpairs:
+            self.hist_logscale = kvpairs['hist_logscale']
+        else:
+            self.hist_logscale = False
         self.format_record_head = kvpairs['format_record_head']
         self.format_record_tail = kvpairs['format_record_tail']
         self.format_snapshot_head = kvpairs['format_snapshot_head']
@@ -1141,10 +1153,23 @@ def set_formats(args, records):
                 break
 
     if fmt.format_snapshot_tail == default_snapshot_tail_format:
+        ops_filters_installed = False
+        intervals_tuning_enabled = False
         for record in records:
             if len(record.scheme_filters) > 0:
-                fmt.format_snapshot_tail = default_snapshot_tail_format_filter_installed
+                ops_filters_installed = True
+            if record.intervals is not None:
+                if record.intervals.intervals_goal.enabled():
+                    intervals_tuning_enabled = True
+            if ops_filters_installed and intervals_tuning_enabled:
                 break
+        if ops_filters_installed:
+            fmt.format_snapshot_tail = default_snapshot_tail_format_filter_installed
+        if intervals_tuning_enabled:
+            fmt.format_snapshot_tail += '\nintervals tuning status: <intervals tuning status>'
+        # further check if scheme action is not stat
+        if args.tried_regions_of is not None:
+            fmt.format_snapshot_tail += '\nscheme stats\n<damos stats>'
 
     if fmt.format_snapshot_head == None:
         need_snapshot_head = False
@@ -1345,6 +1370,8 @@ def add_fmt_args(parser, hide_help=False):
                  'of each region']))
     parser.add_argument('--dont_merge_regions', action='store_true',
             help='don\'t merge contiguous regions of same access pattern')
+    parser.add_argument('--hist_logscale', action='store_true',
+                        help='draw histograms in logscale')
 
     # don't set default for record head and snapshot head because it depends on
     # given number of record and snapshots.  Decide those in set_formats().
